/*				    				      HTEvent.c
**	MULTITHREADED ACCESS MANAGER
**
**	(c) COPYRIGHT MIT 1995.
**	Please first read the full copyright statement in the file COPYRIGH.
**
** Authors:
**	HFN    	Henrik Frystyk Nielsen <frystyk@dxcern.cern.ch>
**
** History:
**      05 Jul 95 Written from scratch
**
** Bugs
**
*/

/* Implementation dependent include files */
#include "tcp.h"

/* Library include files */
#include "HTUtils.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"
#include "HTList.h"
#include "HTEvent.h"					 /* Implemented here */

/* Type definitions and global variables etc. local to this module */
PRIVATE HTList *UserList = NULL;	      /* List of call back functions */
PRIVATE	fd_set	HTfd_user;		   /* Bit array of sockets from user */
extern fd_set HTfd_set ;	  /* all sockets ever set, anytime, anywhere */

#ifdef WIN32
BOOL waitForConsole = NO;
HANDLE cHandle; 
#endif /* WIN32 */

#ifndef _WINDOWS
typedef unsigned long DWORD ;
#endif /* _WINDOWS */

BOOL consoleReady = NO ;


/* --------------------------------------------------------------------------*/


/*							        HTEventRegister
**
**	Registers a socket as active for READ together with a callback
**	function to call when select returns that the socket is pending.
*/
PUBLIC BOOL HTEventRegister ARGS1(HTEventCallBack *, user_fd)
{
    static BOOL done=NO;

    if (!user_fd) {
	if (THD_TRACE)
	    fprintf(TDEST, "EventRegist. Bad argument\n");
	return NO;
    }

    if (!done) {
	FD_ZERO(&HTfd_user);
	done = YES;
    }

    if (!UserList)
	UserList = HTList_new();
    HTList_addObject(UserList, (void *) user_fd);

#ifdef _WIN32 
    if (user_fd->sockfd == STDIN_FILENO) { 
	waitForConsole = YES ;
	cHandle = GetStdHandle( STD_INPUT_HANDLE) ;
    }
    else { 
	/* if its not STDIN, then pretend that it's a socket. select() works on sockets */
#else
    {
#endif

	HTEventThreadState(user_fd->sockfd, THD_SET_READ);  /* CLB bogus if stdin,stdout, stderr */
	FD_SET(user_fd->sockfd, &HTfd_user);
    }
    return YES;
}

PUBLIC BOOL HTEventUnregister ARGS1(HTEventCallBack *, user_fd) 
{
    if (! user_fd ) 
    {
	if (THD_TRACE)
	    fprintf(TDEST, "EventUnregist. Bad argument\n");
	return NO;
    }
    if (UserList != 0)		/* make sure there is one  */
	HTList_removeObject( UserList, (void *)user_fd) ;
    FD_CLR( user_fd -> sockfd, &HTfd_user) ;
    return YES ;
}

/*							         HTEventCleanup
**
**	Clears the list of registered user sockets and call back functions
*/
PUBLIC void HTEventCleanup NOARGS
{
    if (UserList) {
	HTList_delete(UserList);
	UserList = NULL;
    }
}


/*							      HTEventCheckState
**
**	This function checks the return code from the client call back 
**	function HTEventHandler() and HTEventRequestTerminate().
**
**	Returns:
**		YES	If we stay in eventloop
**		NO	If we return to client application
*/
PRIVATE BOOL HTEventCheckState ARGS2(HTRequest *, request, HTEventState, ret)
{
    switch (ret) {
      case EVENT_TERM:					   /* Ignore it here */
      case EVENT_OK:
	break;
	
      case EVENT_INTR:
 	{
	    int sockfd;
	    if (request && request->net_info &&
		(sockfd = request->net_info->sockfd) >= 0)
		HTEventThreadState(sockfd, THD_SET_INTR);
	}
	break;

      case EVENT_INTR_ALL:
	HTThreadMarkIntrAll(&HTfd_user);
	break;
	
      case EVENT_QUIT:
	return NO;
	break;	       
	
      default:
	if (THD_TRACE)
	    fprintf(TDEST, "HTEventLoop. Invalid return code %d\n",ret);
    }
    return YES;
}    


/*								HTEventLoop
**
**	This is the internal socket event loop that is executed when the
**	client doesn't want to keep the event loop. The client can interrupt
**	this loop by sending an event on on of the registered sock descriptors
**	in HTThreadInit() in the HTThread.c module
**
**	The input parameters are for the first load, i.e. the home page.
**	The keywords can be NULL. The Home Anchor can easily be found using
**	HTHomeAnchor().
**
**	Returns:
**		HT_ERROR	if can't get first document or error in select
**				(only if it failed with no blocking)
**		HT_OK		On normal exit
*/
PUBLIC int HTEventLoop ARGS3(HTRequest *,	homerequest,
			     HTParentAnchor *,	homeanchor,
			     CONST char *,	homekeywords)
{
    BOOL user_event = NO;
    int status, selres;
    int cnt;
    HTEventState state;
    HTRequest *cur_request = NULL;
    struct timeval t = {0,0};

    /* Load the Home Page and jump into the event loop */
    status = (homekeywords && *homekeywords) ?
	HTSearch(homekeywords, homeanchor, homerequest) :
	    HTLoadAnchor((HTAnchor *) homeanchor, homerequest);

    if (status != HT_WOULD_BLOCK) {
	if (THD_TRACE)
	    fprintf(TDEST,"HTEventLoop. Calling Terminate\n");
	state = HTEventRequestTerminate(homerequest, status);
	if (!HTEventCheckState(homerequest, state)) {
	    return HT_OK;
	}
    }

    while (1) {					/* Do never jump out of this */
	fd_set fd_read, fd_write;		      /* Set up working copy */
	int maxfdpl;

#ifdef _WIN32
	int i ; 
	DWORD result ;
	DWORD time2wait = 0 ;
	struct timeval t = {1L,0L};
#endif
	FD_ZERO(&fd_read); 
	FD_ZERO(&fd_write) ;
	maxfdpl = HTThreadGetFDInfo(&fd_read, &fd_write);
#ifdef _WIN32

	consoleReady = NO;
	if (waitForConsole) {
	    if (maxfdpl == 0) 	 /* no other sockets live, wait for keyboard */
		time2wait = INFINITE;
	    else 
		time2wait = 0;	   /* this is a poll */
	    switch( result = WaitForSingleObject( cHandle, (DWORD)time2wait)) {
	      case WAIT_TIMEOUT:
		break;
		
	      case WAIT_FAILED :
		case WAIT_ABANDONED: /* should never happen */
		  HTErrorSysAdd(homerequest, ERR_FATAL, GetLastError(), NO,
				"WaitForSingleObject") ;
		return HT_ERROR;
		break ; 
		
	      case WAIT_OBJECT_0:
		FD_SET(STDIN_FILENO, &HTfd_user) ;
		consoleReady = YES ;
		break;
	    } /* switch */
	    if (THD_TRACE)
		fprintf(TDEST, "Console %s ready for input\n",
			consoleReady ? "is" : "isn't" );
	}
#endif /* _WIN32 */

	/*
	 * Charlie Brooks (cbrooks@osf.org)
	 * select with a null timeout will wait forever. We can't do that under NT.
	 * So. We set up a simple timeout of, say, 1 second IFF we have console input.
	 * In effect, we simply poll the other sockets to see if anything is ready at 
	 * this time, but only if console input is available 
	 * 
	 * consoleReady will be YES if we're running NT and the console has IO waiting.
	 * otherwise, it will always be NO.
	 */
#ifdef __hpux
	selres = select(maxfdpl, (int *) &fd_read, (int *) &fd_write,
			(int *) NULL, NULL);
#else
	selres = select(maxfdpl, &fd_read, &fd_write, NULL,
			consoleReady ? &t : NULL);
#endif
	if (THD_TRACE)
	    fprintf(TDEST, "HTEventLoop. Select returns: %d\n", selres);
	
	if (selres < 0) {
	    HTErrorSysAdd(homerequest, ERR_FATAL, socerrno, NO, "select");
	    return HT_ERROR;
	}
	if (selres == 0)
	    goto USER_INPUT;
	
	if (THD_TRACE) {
	    int cnt;
	    fprintf(TDEST, "HTEventLoop. Sockets ready for read : ");
#ifdef _WINSOCKAPI_
	    for (cnt = 0 ; cnt < fd_read.fd_count; cnt++)
		fprintf(TDEST, "%ld ", fd_read.fd_array[cnt]);
#else
	    for (cnt=0; cnt<maxfdpl; cnt++) {
		if (FD_ISSET(cnt, &fd_read))
		    fprintf(TDEST, "%d ", cnt);
	    }
#endif
	    fprintf(TDEST, "\n");
	    fprintf(TDEST, "HTEventLoop. Sockets ready for write: ");
#ifdef _WINSOCKAPI_
	    for (cnt = 0 ; cnt < fd_write.fd_count; cnt++) 
		fprintf(TDEST, "%ld ", fd_write.fd_array[cnt]) ;
#else
	    for (cnt=0; cnt<maxfdpl; cnt++) {
		if (FD_ISSET(cnt, &fd_write))
		    fprintf(TDEST, "%d ", cnt);
	    }
#endif
	    fprintf(TDEST, "\n");
	} /* if THD_TRACE */

	/* Any events on a user socket? - all sockets are examined */  
      USER_INPUT: 
	/* */ ;

#ifdef _WINSOCKAPI_
	for (i = 0 ; i < HTfd_user.fd_count; i++) {
	    cnt = HTfd_user.fd_array[i];
	    if (cnt == STDIN_FILENO) FD_SET(cnt, &fd_read);
#else
	for (cnt=0; cnt<maxfdpl; cnt++) {
#endif
	    if (FD_ISSET(cnt, &fd_read) && FD_ISSET(cnt, &HTfd_user)) {
		HTList *cur = UserList;
		HTEventCallBack *pres;
		while ((pres = (HTEventCallBack *)HTList_nextObject(cur))){
		    if (FD_ISSET(cnt, &HTfd_user)) {
			if (THD_TRACE)
			    fprintf(TDEST,"HTEventLoop. Calling handler %p\n",
				    pres->callback);
			state = (*(pres->callback))(&cur_request);
			if (state == EVENT_TERM)
			    state=HTEventRequestTerminate(cur_request,status);
			if (!HTEventCheckState(cur_request, state))
			    return HT_OK;
			user_event = YES;
		    } /* if */
		} /* while */
	    } /* if */
	}
	if (user_event) {
	    user_event = NO;
#ifdef _WIN32
	    if (FD_ISSET( STDIN_FILENO, &HTfd_user))
		FD_CLR( STDIN_FILENO, &HTfd_user);
#endif	
	    continue; /* Go check for any new User events */
	} else {
	    if (THD_TRACE)
		fprintf(TDEST, "HTEventLoop. No user events found - look for Library events\n");
	}
	    
	/* Any events on a Library socket? - only do this when no user
	   sockets are pending */
	if ((cur_request = HTThread_getRequest(&fd_read, &fd_write)) != NULL) {
	    HTProtocol *dest = (HTProtocol *)
		HTAnchor_protocol(cur_request->anchor);
	    if (dest->load)
		status = (*dest->load)(cur_request);
	    else
		return HT_ERROR;
	    if (THD_TRACE)
		fprintf(TDEST, "HTEventLoop. Callback returns: %d\n", status);
	    if (status != HT_WOULD_BLOCK) {
		HTLoadTerminate(cur_request, status);
		if (THD_TRACE)
		    fprintf(TDEST,"HTEventLoop. Calling Terminate\n");
		state = HTEventRequestTerminate(cur_request, status);
		if (!HTEventCheckState(cur_request, state)) {
		    return HT_OK;
		}
	    }
	} else {
	    if (THD_TRACE)
		fprintf(TDEST, "HTEventLoop. No library events found\n");
	}
    } /* End of while */
    return HT_ERROR;				    /* Should never get here */
}

